! ################################################
! # Compute the model coefficient c0 as a function
! # of gamma and Ret according to ARM model
! #
! # Definition of model spectrum
! # 
! # p(1) - p    (low frequency exponent)
! # p(2) - beta (5.2 for Pope)
! # p(3) - cl
! # p(4) - ret
! # p(5) - cn
! # p(6) - cK   (1.5 for Pope)
! # p(7) - m    (slope, 5/3 for Pope)
! # p(8) - n    (viscous exponent, 1 for Pope)
! # 
! ################################################

      SUBROUTINE SGS_ARM_TABLE_P(ng, remin, dre, nre, vg, vret, c0)

      IMPLICIT NONE

#include "types.h"
#include "nfile.h"
#include "dns_error.h"
      
      TINTEGER ng, nre
      TREAL remin, dre
      TREAL c0(ng, nre)
      TREAL vg(ng, nre)
      TREAL vret(nre)

      TREAL gamma, rg

      TREAL params(9)
      TREAL cl, cn, ret, cl0, cn0
      TREAL A0, A1, A2, c34
      TINTEGER niter, nmax, ire, ig
      
      TREAL A,B,EPSABS
      TINTEGER IER

      TREAL ESP_P_CLCN_ROOT, ESP_P_A0, ESP_P_A1, ESP_P_A2
      EXTERNAL ESP_P_CLCN_ROOT, ESP_P_A0, ESP_P_A1, ESP_P_A2
      CHARACTER*64 message

      c34 = C_3_R/C_4_R
      message = 'SGS_ARM_TABLE_P : Error in integration IER = '

! Integration and Iteration constants
      nmax = 128
      EPSABS = C_1EM8_R

! Parameters defining the spectrum
#ifdef SINGLE_PREC
      params(1) = C_2_R
      params(2) = 5.2e0*1.19e0
      params(6) = 0.67e0
      params(7) = C_5_R/C_3_R
      params(8) = C_1_R
#else
      params(1) = C_2_R
      params(2) = 5.2d0*1.19d0
      params(6) = 0.67d0
      params(7) = C_5_R/C_3_R
      params(8) = C_1_R
#endif

! Lower integration limit
      A = C_0_R

! Initialization Parameters for Newton-Raphson. They are ok for low Ret
      cl0 = C_05_R
      cn0 = C_05_R

! Loop over the Reynolds number
      ret = remin
      DO ire = 1,nre

! Set reynolds table to log(ret)
         vret(ire) = LOG(ret)

! turbulent reynolds number
         params(4) = ret

! Find cl(Ret) function so k and epsilon are exact.
         CALL NEWTON2(ESP_P_CLCN_ROOT, params, cl0, cn0, nmax, EPSABS, &
              niter, cl, cn)
#ifdef _DEBUG
         WRITE(*,*) niter, cl, cn
#endif
         params(3) = cl
         params(5) = cn

! Set next cl0 and cn0 to the found root to accelerate convergence
         cl0 = cl
         cn0 = cn

! Minimum gamma is eta/2 L = (Re**-3/4)/2
         gamma = C_05_R/(ret**c34)

! Gamma factor is set so last gamma = 1/2 (Delta_f == L)
         rg = EXP(c34*LOG(ret)/M_REAL(ng-1))
         DO ig = 1, ng

! Set integral scale array to log
            vg(ig,ire) = LOG(gamma)

! Gamma parameter of model
            params(9) = gamma

! Calculate c0(Ret,gama)
            B = 1/gamma
            CALL INTEGSPC(ESP_P_A0,params,A,B,EPSABS,A0,IER)
            IF (IER.NE.0) THEN
               CALL CONCATI(message,IER)
               CALL IO_WRITE_ASCII(efile,message)
               CALL DNS_STOP(LES_ERROR_ARM_INI)
            ENDIF
            CALL INTEGSPC(ESP_P_A1,params,A,B,EPSABS,A1,IER)
            IF (IER.NE.0) THEN
               CALL CONCATI(message,IER)
               CALL IO_WRITE_ASCII(efile,message)
               CALL DNS_STOP(LES_ERROR_ARM_INI)
            ENDIF
            CALL INTEGSPC(ESP_P_A2,params,A,B,EPSABS,A2,IER)
            IF (IER.NE.0) THEN
               CALL CONCATI(message,IER)
               CALL IO_WRITE_ASCII(efile,message)
               CALL DNS_STOP(LES_ERROR_ARM_INI)
            ENDIF
            
! Calculate c0 from root of quadratic equation
            c0(ig,ire) = (-A1+SQRT(A1**2-C_4_R*A0*A2))*C_05_R/A2

! Increase gamma
            gamma = gamma*rg
         ENDDO

! Increase Ret
         ret = ret*dre
      ENDDO

      RETURN
      END
      
