!########################################################################
!# Tool/Library FIELDS
!#
!########################################################################
!# HISTORY
!#
!# 2007/09/12 - J.P. Mellado
!#              Created
!#
!########################################################################
!# DESCRIPTION
!#
!# Calculate the strain production term as given by -s_ij s_jk s_ki 
!# minus 1/4 of the vorticity production. The first term is:
!# s_11^3 + s_22^3 +  s_33^3 +
!# 3 s_11 (s_12^2+s_13^2) + 3 s_22 (s_21^2+s_23^2) + 3 s_33 (s_31^2+s_32^2) +
!# 2 s_12 s_23 s_31
!# or
!# 2 s_12 s_23 s_31 +
!# s_11 ( s_11^2 + 3 ( s_12^2 + s_13^2 ) ) +
!# s_22 ( s_22^2 + 3 ( s_12^2 + s_23^2 ) ) +
!# s_33 ( s_33^2 + 3 ( s_13^2 + s_23^2 ) )
!########################################################################
!# ARGUMENTS 
!#
!########################################################################
SUBROUTINE FI_STRAIN_PRODUCTION&
     (imode_fdm, imax, jmax, kmax, i1bc, j1bc, k1bc, &
     dx, dy, dz, u, v, w, result, s_12, s_13, s_23, tmp1, tmp2, &
     wrk1d, wrk2d, wrk3d)

  IMPLICIT NONE

#include "types.h"
#include "integers.h"

  TINTEGER imode_fdm, imax, jmax, kmax, i1bc, j1bc, k1bc

  TREAL dx(imax)
  TREAL dy(jmax)
  TREAL dz(kmax)

  TREAL u(*), v(*), w(*), result(*)
  TREAL s_12(*), s_13(*), s_23(*)
  TREAL tmp1(*), tmp2(*)

  TREAL wrk1d(imax,*)
  TREAL wrk2d(imax,*)
  TREAL wrk3d(imax,jmax,kmax)

  ! -------------------------------------------------------------------
  TINTEGER ij

  ! ###################################################################
  ! Vorticity production part
  ! ###################################################################
  CALL FI_VORTICITY_PRODUCTION&
       (imode_fdm, imax, jmax, kmax, i1bc, j1bc, k1bc, &
       dx, dy, dz, u, v, w, result, s_12, s_13, s_23, tmp1, tmp2, &
       wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     result(ij) = C_025_R*result(ij)
  ENDDO

  ! ###################################################################
  ! Pure strain term
  ! ###################################################################
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, v, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, u, tmp2, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     s_12(ij) = C_05_R*(tmp1(ij)+tmp2(ij))
  ENDDO

  ! Uz, Wx
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, w, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, u, tmp2, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     s_13(ij) = C_05_R*(tmp1(ij)+tmp2(ij))
  ENDDO

  ! Vz, Wy
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, w, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, v, tmp2, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     s_23(ij) = C_05_R*(tmp1(ij)+tmp2(ij))
  ENDDO

  ! -------------------------------------------------------------------
  ! Production term
  ! -------------------------------------------------------------------
  DO ij = 1,imax*jmax*kmax
     result(ij) = result(ij) + C_2_R*s_12(ij)*s_13(ij)*s_23(ij)
  ENDDO

  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, u, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     result(ij) = result(ij) + tmp1(ij)*(&
          tmp1(ij)*tmp1(ij) + C_3_R*(s_12(ij)*s_12(ij)+s_13(ij)*s_13(ij)) )
  ENDDO

  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, v, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     result(ij) = result(ij) + tmp1(ij)*(&
          tmp1(ij)*tmp1(ij) + C_3_R*(s_12(ij)*s_12(ij)+s_23(ij)*s_23(ij)) )
  ENDDO

  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, w, tmp1, i0, i0, wrk1d, wrk2d, wrk3d)
  DO ij = 1,imax*jmax*kmax
     result(ij) = result(ij) + tmp1(ij)*(&
          tmp1(ij)*tmp1(ij) + C_3_R*(s_13(ij)*s_13(ij)+s_23(ij)*s_23(ij)) )
  ENDDO

  ! right sign
  DO ij = 1,imax*jmax*kmax
     result(ij) =-result(ij)
  ENDDO

  RETURN
END SUBROUTINE FI_STRAIN_PRODUCTION
