!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2001/11/30 - J.P. Mellado
!#              Created
!# 2008/11/16 - J.P. Mellado
!#              New code version based on energy
!#
!########################################################################
!# DESCRIPTION
!#
!# This subroutine computes the additive Smagorinsky part of the
!# residual stress term of the right
!# hand side of the filtered compressible Navier-Stokes equations using a 
!# dynamic model.
!# It is computationally very expensive. Some operations were duplicated 
!# to avoid the use of additional arrays. However, txc had to be enlarged
!# by 1.
!# Filter level 0 = n0
!# Filter level 1 = n1  (test filter)
!#
!# FAVRE_AVG not used to save filter ops. and div. of rho     
!#
!########################################################################
!# ARGUMENTS 
!#
!# txc    In    size of 10 for dynamic SMG
!#              size of 13 for dynamic mixed
!#
!########################################################################
SUBROUTINE SGS_DYN(istep, statflg, ssmflg, x,y,z,dx,dy,dz, q,z1, &
     u_sg,v_sg,w_sg,aux_sg, h1,h2,h3,h4,zh1, txc, mean1d_les, chi, wrk1d,wrk2d,wrk3d, &
     cf0x,cf0y,cf0z,cf1x,cf1y,cf1z,cf2x,cf2y,cf2z, &
     carm_uf,carm_uh,carm_pf,carm_ph,carm_zf,carm_zh, sgscu,sgsce,sgscz)

#include "types.h"
#include "dns_const.h"
#include "les_const.h"
#include "les_avgij_map.h"

  USE DNS_GLOBAL
#ifdef PARALLEL
  USE DNS_MPI
#endif
  USE LES_GLOBAL

  IMPLICIT NONE

#include "integers.h"
#ifdef PARALLEL
#include "mpif.h"
#endif

  TINTEGER istep, statflg, ssmflg

  TREAL, DIMENSION(*)                :: x, y, z, dx, dy, dz
  TREAL, DIMENSION(imax,jmax,kmax)   :: u_sg, v_sg, w_sg, aux_sg, h1, h2, h3, h4, chi, wrk3d
  TREAL, DIMENSION(imax,jmax,kmax,*) :: q, z1, zh1, txc
  TREAL, DIMENSION(nstatavg,jmax,*)  :: mean1d_les
  TREAL, DIMENSION(*)                :: wrk1d, wrk2d

  TREAL, DIMENSION(*) :: cf0x, cf0y, cf0z, cf1x, cf1y, cf1z, cf2x, cf2y, cf2z
  TREAL, DIMENSION(*) :: carm_uf, carm_uh, carm_pf, carm_ph, carm_zf, carm_zh
  TREAL, DIMENSION(*) :: sgscu, sgsce, sgscz

  TARGET :: q

! -----------------------------------------------------------------------
  TREAL coefdelta
  TINTEGER ip_den, ip_tmp, ip_tmp2, ip_loc, i, j, k, ij
  TINTEGER n1, n0, is, statflg_loc

! Pointers to existing allocated space
  TREAL, DIMENSION(:,:,:),   POINTER :: u, v, w, e, rho

#ifdef PARALLEL
  TREAL dummy
#endif

! #######################################################################
#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'ENTERING SGS_DYN' )
#endif

! Define pointers
  u   => q(:,:,:,1)
  v   => q(:,:,:,2)
  w   => q(:,:,:,3)
  e   => q(:,:,:,4)
  rho => q(:,:,:,5)

  n0 = isgs_f0size
  n1 = isgs_f1size
  coefdelta = (M_REAL(n1)/M_REAL(n0))**C_2_R + C_1_R

! defining the array distribution
  ip_den = 3
  IF ( ssmflg .EQ. 1 ) THEN
     ip_tmp = ip_den + 3
  ELSE
     ip_tmp = ip_den + 1
  ENDIF
  ip_tmp2 = ip_tmp + 1

#define STRAIN0(i,j,k)        txc(i,j,k,1)
#define STRAIN1(i,j,k)        txc(i,j,k,2)
#define RHOINV1(i,j,k)        txc(i,j,k,ip_den)

! #######################################################################
! Calculating common arrays
! #######################################################################
! -----------------------------------------------------------------------
! specific volumes
! -----------------------------------------------------------------------
! txc_size=1
  CALL SGS_DYN_BARVS(n0, n1, rho, RHOINV1(1,1,1), txc(1,1,1,ip_tmp), wrk3d, &
       cf0x, cf0y, cf0z, cf1x, cf1y, cf1z, cf2x, cf2y, cf2z)

! -----------------------------------------------------------------------
! Define the base fields over which the dynamic procedure 
! applies (i.e. remove mean if so specified)
! -----------------------------------------------------------------------
! txc_size=1
  CALL SGS_FIELDS(iles_type_regu, n0, dx, dy, dz, &
       cf0x, cf0y, cf0z, rho, u, u_sg, txc(1,1,1,ip_tmp), wrk2d, wrk3d)
  CALL SGS_FIELDS(iles_type_regu, n0, dx, dy, dz, &
       cf0x, cf0y, cf0z, rho, v, v_sg, txc(1,1,1,ip_tmp), wrk2d, wrk3d)
  CALL SGS_FIELDS(iles_type_regu, n0, dx, dy, dz, &
       cf0x, cf0y, cf0z, rho, w, w_sg, txc(1,1,1,ip_tmp), wrk2d, wrk3d)

! -----------------------------------------------------------------------
! ARM coefficient
! -----------------------------------------------------------------------
! txc_size=6
!  IF ( iles_type_tran .EQ. LES_TRAN_ARMRANS .OR.&
!       iles_type_tran .EQ. LES_TRAN_ARMSPTL .OR.&
!       iles_type_tran .EQ. LES_TRAN_ARMINVS ) THEN
! not yet implemented in the new version
!     CALL SGS_DYN_ARM_U(n0, n1, x, y, z, dx, dy, dz, &
!          rho, u_sg, v_sg, w_sg, RHOINV1(1,1,1), &
!          txc(1,1,1,ip_tmp), wrk1d, wrk2d, wrk3d, &
!          cf0x, cf0y, cf0z,&
!          cf1x, cf1y, cf1z, cf2x, cf2y, cf2z, &
!          carm_uf, carm_uh)
!  ENDIF

! -----------------------------------------------------------------------
! Norms of strain rate tensor
! -----------------------------------------------------------------------
! txc_size=6
  CALL SGS_DYN_SIJ(n0, n1, x, y, z, dx, dy, dz,&
       rho, u_sg, v_sg, w_sg, STRAIN0(1,1,1), STRAIN1(1,1,1), &
       RHOINV1(1,1,1), txc(1,1,1,ip_tmp), wrk1d, wrk2d, wrk3d, cf1x, cf1y, cf1z)

! #######################################################################
! Velocity
! #######################################################################
! ------------------------------------------------------------
!          Smagorinsky dynamic coefficient (txc_size=7or8)
! ------------------------------------------------------------
! imediately after SGS_DYN_SIJ due to dilatation calculation
  IF ( istep .EQ. 1 ) THEN
     CALL SGS_DYN_U(ssmflg, coefdelta,n0,n1, x,y,z,dx,dy,dz, rho,u,v,w, u_sg,v_sg,w_sg, &
          STRAIN0(1,1,1), STRAIN1(1,1,1), RHOINV1(1,1,1), txc(1,1,1,ip_tmp), &
          sgscu, wrk1d,wrk2d,wrk3d, &
          cf0x,cf0y,cf0z,cf1x,cf1y,cf1z,cf2x,cf2y,cf2z, carm_uf,carm_uh)
  ENDIF

! -----------------------------------------------------------------------
! txc(ip_tmp) is 2*rho*nu_s = 2*rho*dyn_coef*sgs_delta^2*S
  IF ( imode_sim .EQ. DNS_MODE_TEMPORAL ) THEN
     smg_ucoef = C_05_R*MAXVAL(sgscu(1:jmax))
     DO k = 1,kmax
        DO j = 1,jmax
           DO i = 1,imax
              txc(i,j,k,ip_tmp) = sgscu(j)*STRAIN0(i,j,k)
           ENDDO
        ENDDO
     ENDDO

  ELSE IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
     smg_ucoef = C_05_R*MAXVAL(sgscu(1:imax*jmax))
     DO k = 1,kmax
        DO ij = 1,imax*jmax
           txc(ij,1,k,ip_tmp) = sgscu(ij)*STRAIN0(ij,1,k)
        ENDDO
     ENDDO
! Running statistics
     IF ( statflg .EQ. 1 ) THEN
        CALL LES_SAVE_AVGIJ(i1, dx,dy,dz, u,v,w, txc(1,1,1,ip_tmp), mean1d_les, &
             wrk1d, sgscu, wrk3d)
     ENDIF

  ENDIF

! Obtain maximum nu_s for D# restriccion. 
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        wrk3d(ij,1,k) = txc(ij,1,k,ip_tmp)/rho(ij,1,k)
     ENDDO
  ENDDO
  smg_udiff = MAXVAL(wrk3d)
#ifdef PARALLEL
  CALL MPI_ALLREDUCE(smg_udiff, dummy, 1, MPI_REAL8, MPI_MAX, MPI_COMM_WORLD, ims_err)
  smg_udiff = dummy
#endif

! ------------------------------------------------------------
! Transport terms (txc_size=6)
! ------------------------------------------------------------
! 2*rho*nu_sg in txc(ip_tmp)
  CALL SGS_SMG_U(statflg, i0, dx,dy,dz, rho, u_sg,v_sg,w_sg, u,v,w, &
       txc(1,1,1,ip_tmp), h1,h2,h3,h4, &
       txc(1,1,1,ip_tmp2), mean1d_les, wrk1d,wrk2d,wrk3d)

! #######################################################################
! Energy
! #######################################################################
! ------------------------------------------------------------
!           Smagorinsky dynamic coefficient (txc_size=6or7)
! ------------------------------------------------------------
  IF ( istep .EQ. 1 ) THEN
     CALL SGS_DYN_Z(ssmflg, coefdelta,n0,n1, x,y,z,dx,dy,dz, rho,u,v,w,e, u_sg,v_sg,w_sg,aux_sg,&
          STRAIN0(1,1,1), STRAIN1(1,1,1), RHOINV1(1,1,1), txc(1,1,1,ip_tmp), &
          sgsce, wrk1d,wrk2d,wrk3d, &
          cf0x,cf0y,cf0z,cf1x,cf1y,cf1z,cf2x,cf2y,cf2z, carm_uf,carm_uh,carm_pf,carm_ph)
  ENDIF

! -----------------------------------------------------------------------
! txc(ip_tmp) is rho*nu_s = rho*CsE*sgs_delta^2*S
  IF ( imode_sim .EQ. DNS_MODE_TEMPORAL ) THEN
     smg_ecoef = MAXVAL(sgsce(1:jmax))
     DO k = 1,kmax
        DO j = 1,jmax
           DO i = 1,imax
              txc(i,j,k,ip_tmp) = sgsce(j)*STRAIN0(i,j,k)
           ENDDO
        ENDDO
     ENDDO

  ELSE IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
     smg_ecoef = MAXVAL(sgsce(1:imax*jmax))
     DO k = 1,kmax
        DO ij = 1,imax*jmax
           txc(ij,1,k,ip_tmp) = sgsce(ij)*STRAIN0(ij,1,k)
        ENDDO
     ENDDO
! Running statistics
     IF ( statflg .EQ. 1 ) THEN
        CALL LES_SAVE_AVGPIJ(i1, dx,dy,dz, e, txc(ij,1,k,ip_tmp), mean1d_les, &
             wrk1d, sgsce, wrk3d)
     ENDIF

  ENDIF

! Obtain maximum nu_s for D# restriccion. 
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        wrk3d(ij,1,k) = txc(ij,1,k,ip_tmp)/rho(ij,1,k)
     ENDDO
  ENDDO
  smg_ediff = MAXVAL(wrk3d)
#ifdef PARALLEL
  CALL MPI_ALLREDUCE(smg_ediff, dummy, 1, MPI_REAL8, MPI_MAX, MPI_COMM_WORLD, ims_err)
  smg_ediff = dummy
#endif

! ------------------------------------------------------------
! Transport terms (txc_size=2)
! ------------------------------------------------------------
  statflg_loc =-statflg
  CALL SGS_SMG_Z(statflg_loc, dx,dy,dz, rho, e, aux_sg, &
       txc(1,1,1,ip_tmp), h4, &
       mean1d_les, &
       txc(1,1,1,ip_tmp2), chi, wrk1d, wrk2d, wrk3d)

! #######################################################################
! Scalar
! #######################################################################
  IF ( icalc_scal .EQ. 1 ) THEN
     DO is = 1,inb_scal
        IF ( imode_sim .EQ. DNS_MODE_TEMPORAL ) THEN
           ip_loc = 1 + (is-1)*jmax
        ELSE IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
           ip_loc = 1 + (is-1)*imax*jmax
        ENDIF
! ------------------------------------------------------------
!           Smagorinsky dynamic coefficient (txc_size=6or7)
! ------------------------------------------------------------
        IF ( istep .EQ. 1 ) THEN
           CALL SGS_DYN_Z(ssmflg, coefdelta,n0,n1, x,y,z,dx,dy,dz, &
                rho,u,v,w,z1(1,1,1,is), u_sg,v_sg,w_sg,aux_sg, &
                STRAIN0(1,1,1), STRAIN1(1,1,1), RHOINV1(1,1,1), txc(1,1,1,ip_tmp), &
                sgscz(ip_loc), wrk1d,wrk2d,wrk3d, &
                cf0x,cf0y,cf0z,cf1x,cf1y,cf1z,cf2x,cf2y,cf2z, carm_uf,carm_uh,carm_zf,carm_zh)
        ENDIF

! -----------------------------------------------------------------------
! txc(ip_tmp) is rho*nu_s = rho*CsZ*sgs_delta^2*S
        IF ( imode_sim .EQ. DNS_MODE_TEMPORAL ) THEN
           smg_zcoef = MAXVAL(sgscz(ip_loc:ip_loc+jmax-1))
           DO k = 1,kmax
              DO j = 1,jmax
                 DO i = 1,imax
                    txc(i,j,k,ip_tmp) = sgscz(j+ip_loc-1)*STRAIN0(i,j,k)
                 ENDDO
              ENDDO
           ENDDO
           
        ELSE IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
           smg_zcoef = MAXVAL(sgscz(ip_loc:ip_loc+imax*jmax-1))
           DO k = 1,kmax
              DO ij = 1,imax*jmax
                 txc(ij,1,k,ip_tmp) = sgscz(ij+ip_loc-1)*STRAIN0(ij,1,k)
              ENDDO
           ENDDO
! Running statistics
           IF ( statflg .EQ. 1 ) THEN
              CALL LES_SAVE_SCBDGIJ(i1, dx,dy,dz, z1, txc(1,1,1,ip_tmp), &
                   mean1d_les(1,1,1+LA_MOMENTUM_SIZE+(is-1)*LS_SCALAR_SIZE), &
                   wrk1d, sgscz(ip_loc), wrk3d)
           ENDIF

        ENDIF

! Obtain maximum nu_s for D# restriccion. The maximum over the PEs is done in courant.f
! but we need it for dns.les  
        DO k = 1,kmax
           DO ij = 1,imax*jmax
              wrk3d(ij,1,k) = txc(ij,1,k,ip_tmp)/rho(ij,1,k)
           ENDDO
        ENDDO
        smg_zdiff = MAXVAL(wrk3d)
#ifdef PARALLEL
        CALL MPI_ALLREDUCE(smg_zdiff, dummy, 1, MPI_REAL8, MPI_MAX, MPI_COMM_WORLD, ims_err)
        smg_zdiff = dummy
#endif

! ------------------------------------------------------------
! Transport terms
! ------------------------------------------------------------
        CALL SGS_SMG_Z(statflg, dx,dy,dz, rho, z1(1,1,1,is), aux_sg, &
             txc(1,1,1,ip_tmp), zh1(1,1,1,is), &
             mean1d_les(1,1,1+LA_MOMENTUM_SIZE+(is-1)*LS_SCALAR_SIZE), &
             txc(1,1,1,ip_tmp2), chi, wrk1d,wrk2d,wrk3d)
     ENDDO
  ENDIF

#undef STRAIN0
#undef STRAIN1
#undef RHOINV1

#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'LEAVING SGS_DYN' )
#endif

  RETURN
END SUBROUTINE SGS_DYN

