!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2001/11/30 - J.P. Mellado
!#              Created
!# 12/03/01 - statflg and mean1d_les added for statistics
!# 01/22/02 - extracted from sgs_dynamic.f    
!# 09/29/04 - Introducing dissipation terms
!# 09/03/05 - Introducing viscous part here to reduce CPU time 
!#
!########################################################################
!# DESCRIPTION
!#
!# Add SGS terms to the energy and momentum equations using Smagorinsky
!# model (dynamic or static goes into the viscosity)
!#
!########################################################################
!# ARGUMENTS 
!#
!# statflg    In  use running statictics
!# staticflg  In  coming from sgs_static, shear strains are already there
!# sgsvis     In  Array for 2*\mu_sg
!# txc        In  size 6
!#
!########################################################################
SUBROUTINE SGS_SMG_U(statflg, staticflg, dx, dy, dz, rho, u, v, w, u_r, v_r, w_r, &
     sgsvis, h1, h2, h3, h4, txc, mean1d_les, wrk1d, wrk2d, wrk3d)

#include "types.h"
#include "dns_const.h"
#include "les_const.h"

  USE DNS_GLOBAL
  USE LES_GLOBAL

  IMPLICIT NONE

#include "integers.h"

  TINTEGER statflg, staticflg

  TREAL, DIMENSION(*)                :: dx, dy, dz
  TREAL, DIMENSION(imax,jmax,kmax)   :: u, v, w, rho, u_r, v_r, w_r, sgsvis, h1, h2, h3, h4
  TREAL, DIMENSION(imax,jmax,kmax,*) :: txc

  TREAL mean1d_les(nstatavg,jmax,*)

  TREAL, DIMENSION(*) :: wrk1d, wrk2d, wrk3d

! -----------------------------------------------------------------------
  TINTEGER ij, k
  TREAL r23, r13
  TINTEGER i1vsin, i1vsout, imxvsin, imxvsout
  TINTEGER j1vsin, j1vsout, jmxvsin, jmxvsout
  TINTEGER k1vsin, k1vsout, kmxvsin, kmxvsout

  r23 = C_2_R/C_3_R
  r13 = C_1_R/C_3_R

! #######################################################################
#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'ENTERING SGS_SMG_U' )
#endif

#include "dns_bcs_inf.h"
#include "dns_bcs_out.h"

! #######################################################################
! Off-diagonal terms of SGS stress tensor
! #######################################################################
#define SXY(i,j,k)       txc(i,j,k,1)
#define SXZ(i,j,k)       txc(i,j,k,2)
#define SYZ(i,j,k)       txc(i,j,k,3)
#define TMP_LOC(i,j,k)   txc(i,j,k,4)

  IF ( staticflg .EQ. 0 ) THEN
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, u, SXY(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, v, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           SXY(ij,1,k) = C_05_R*(SXY(ij,1,k)+TMP_LOC(ij,1,k))
        ENDDO
     ENDDO
     CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
          dz, u, SXZ(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, w, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           SXZ(ij,1,k) = C_05_R*(SXZ(ij,1,k)+TMP_LOC(ij,1,k))
        ENDDO
     ENDDO
     CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
          dz, v, SYZ(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, w, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           SYZ(ij,1,k) = C_05_R*(SYZ(ij,1,k)+TMP_LOC(ij,1,k))
        ENDDO
     ENDDO

  ENDIF
#undef TMP_LOC

! Keep this order of Sxy, Sxz, Syz and Tau_smg_m for LES_SAVE_AVGIJ !!!
#define TAU_M(i,j,k)     txc(i,j,k,4)
#define TMP_LOC(i,j,k)   txc(i,j,k,5)

! -----------------------------------------------------------------    
!                                            tauxy = - mu_s*Sxy      
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
  DO k=1,kmax
     DO ij=1,imax*jmax
        TAU_M(ij,1,k) = -SXY(ij,1,k)*sgsvis(ij,1,k)
     ENDDO
  ENDDO
  IF ( statflg .EQ. 1 ) THEN
     CALL LES_SAVE_AVGIJ(i2, dx, dy, dz, u, v, w,&
          SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

! Dissipation, - (tau_xy*du/dy+tau_yx*dv/dx)
  IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, u_r, SXY(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, v_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k = 1,kmax
        DO ij = 1,imax*jmax
           h4(ij,1,k) = h4(ij,1,k) -  TAU_M(ij,1,k)*(SXY(ij,1,k)+TMP_LOC(ij,1,k))
        ENDDO
     ENDDO
  ENDIF

! Calculate  - d/dy(tauxy)             (X-momentum)
!       and  - d/dx(tauyx)             (Y-momentum)
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, TAU_M(1,1,1), SXY(1,1,1), j1vsout, jmxvsout,&
       wrk1d, wrk2d, wrk3d)      
  DO k=1,kmax
     DO ij=1,imax*jmax
        h1(ij,1,k) = h1(ij,1,k) - SXY(ij,1,k)
     ENDDO
  ENDDO
  IF ( statflg .EQ. 1 ) THEN
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, TAU_M(1,1,1), SXY(1,1,1), j1vsout, jmxvsout,&
          wrk1d, wrk2d, wrk3d)
     CALL LES_SAVE_AVGIJ(i14, dx, dy, dz, u, v, w,&
          SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, TAU_M(1,1,1), SXY(1,1,1), i1vsout, imxvsout,&
       wrk1d, wrk2d, wrk3d)  
  DO k=1,kmax
     DO ij=1,imax*jmax
        h2(ij,1,k) = h2(ij,1,k) - SXY(ij,1,k)
     ENDDO
  ENDDO
  IF ( statflg .EQ. 1 ) THEN
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, TAU_M(1,1,1), SXY(1,1,1), i1vsout, imxvsout,&
          wrk1d, wrk2d, wrk3d)  
     CALL LES_SAVE_AVGIJ(i15, dx, dy, dz, u, v, w,&
          SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

  IF ( kmax_total .GT. 1 ) THEN
! -----------------------------------------------------------------    
!                                            tauxz = - mu_s*Sxz      
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -SXZ(ij,1,k)*sgsvis(ij,1,k)
        ENDDO
     ENDDO
     IF ( statflg .EQ. 1 ) THEN
! !!!! yes, SXY
        CALL LES_SAVE_AVGIJ(i3, dx, dy, dz, u, v, w,&
             SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d )
     ENDIF

! Dissipation, - (tau_xz*du/dz+tau_zx*dw/dx)
     IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, u_r, SXZ(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
        CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
             dx, w_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
        DO k = 1,kmax
           DO ij = 1,imax*jmax
              h4(ij,1,k) = h4(ij,1,k) - TAU_M(ij,1,k)*(SXZ(ij,1,k)+TMP_LOC(ij,1,k))
           ENDDO
        ENDDO
     ENDIF

! Calculate  - d/dz(tauxz)             (X-momentum)
!       and  - d/dx(tauzx)             (Z-momentum)
     CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
          dz, TAU_M(1,1,1), SXY(1,1,1), k1vsout, kmxvsout,&
          wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, TAU_M(1,1,1), SXZ(1,1,1), i1vsout, imxvsout,&
          wrk1d, wrk2d, wrk3d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           h1(ij,1,k) = h1(ij,1,k) - SXY(ij,1,k)
           h3(ij,1,k) = h3(ij,1,k) - SXZ(ij,1,k)
        ENDDO
     ENDDO
     IF ( statflg .EQ. 1 ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, TAU_M(1,1,1), SXY(1,1,1), k1vsout, kmxvsout,&
             wrk1d, wrk2d, wrk3d)
        CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
             dx, TAU_M(1,1,1), SXZ(1,1,1), i1vsout, imxvsout,&
             wrk1d, wrk2d, wrk3d)
        CALL LES_SAVE_AVGIJ(i18, dx, dy, dz, u, v, w,&
             SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     ENDIF

! -----------------------------------------------------------------    
!                                            tauyz = - mu_s*Syz      
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -SYZ(ij,1,k)*sgsvis(ij,1,k)
        ENDDO
     ENDDO
     IF ( statflg .EQ. 1 ) THEN
! ! Yes, Sxy
        CALL LES_SAVE_AVGIJ(i4, dx, dy, dz, u, v, w,&
             SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     ENDIF

! Dissipation, - (tau_yz*dv/dz+tau_zy*dw/dy)
     IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, v_r, SYZ(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
        CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
             dy, w_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
        DO k = 1,kmax
           DO ij = 1,imax*jmax
              h4(ij,1,k) = h4(ij,1,k) - TAU_M(ij,1,k)*(SYZ(ij,1,k)+TMP_LOC(ij,1,k))
           ENDDO
        ENDDO
     ENDIF

! Calculate  - d/dz(tauyz)             (Y-momentum)
!       and  - d/dy(tauzy)             (Z-momentum)
     CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
          dz, TAU_M(1,1,1), SXY(1,1,1), k1vsout, kmxvsout,&
          wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, TAU_M(1,1,1), SXZ(1,1,1), j1vsout, jmxvsout,&
          wrk1d, wrk2d, wrk3d)           
     DO k=1,kmax
        DO ij=1,imax*jmax
           h2(ij,1,k) = h2(ij,1,k) - SXY(ij,1,k)
           h3(ij,1,k) = h3(ij,1,k) - SXZ(ij,1,k)
        ENDDO
     ENDDO
     IF ( statflg .EQ. 1 ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, TAU_M(1,1,1), SXY(1,1,1), k1vsout, kmxvsout,&
             wrk1d, wrk2d, wrk3d)
        CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
             dy, TAU_M(1,1,1), SXZ(1,1,1), j1vsout, jmxvsout,&
             wrk1d, wrk2d, wrk3d)
        CALL LES_SAVE_AVGIJ(i19, dx, dy, dz, u, v, w,&
             SXY(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     ENDIF

  ENDIF

#undef SXY
#undef SXZ
#undef SYZ
#undef TAU_M
#undef TMP_LOC

! #######################################################################
! Diagonal terms of the SGS stress tensor
! #######################################################################
#define UXX(i,j,k)       txc(i,j,k,1)
#define UYY(i,j,k)       txc(i,j,k,2)
#define UZZ(i,j,k)       txc(i,j,k,3)
#define TAU_M(i,j,k)     txc(i,j,k,4)
#define TMP_LOC(i,j,k)   txc(i,j,k,5)
! !!! Keep this order in txc1-txc5 for statistics !!!
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, u, UXX(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, v, UYY(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, w, UZZ(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)

! -----------------------------------------------------------------    
!                                            tauxx = - mu_s*Sxx   
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
  IF ( sgs_devsmag .EQ. 1 ) THEN
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                (r23*UXX(ij,1,k)-&
                r13*(UYY(ij,1,k)+UZZ(ij,1,k)))
        ENDDO
     ENDDO
  ELSE
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                UXX(ij,1,k)
        ENDDO
     ENDDO
  ENDIF
! Running statistics
  IF ( statflg .EQ. 1 ) THEN
     CALL LES_SAVE_AVGIJ(i5, dx, dy, dz, u, v, w,&
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

! Dissipation, -tau_xx*du/dx
  IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, u_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k = 1,kmax
        DO ij = 1,imax*jmax
           h4(ij,1,k) =  h4(ij,1,k) - TAU_M(ij,1,k)*TMP_LOC(ij,1,k)
        ENDDO
     ENDDO
  ENDIF

! Calculate  - d/dx(tauxx)             (X-momentum)
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, TAU_M(1,1,1), TMP_LOC(1,1,1), i1vsin, imxvsin,&
       wrk1d, wrk2d, wrk3d)     
  DO k=1,kmax
     DO ij=1,imax*jmax
        h1(ij,1,k) = h1(ij,1,k) - TMP_LOC(ij,1,k)
     ENDDO
  ENDDO
! Running statistics
  IF ( statflg .EQ. 1 ) THEN
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, TAU_M(1,1,1), TMP_LOC(1,1,1), i1vsin, imxvsin,&
          wrk1d, wrk2d, wrk3d)
     CALL LES_SAVE_AVGIJ(i16, dx, dy, dz, u, v, w,&
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, TAU_M(1,1,1), TMP_LOC(1,1,1), j1vsin, jmxvsin,&
          wrk1d, wrk2d, wrk3d) 
     CALL LES_SAVE_AVGIJ(i27, dx, dy, dz, u, v, w,&
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

! -----------------------------------------------------------------    
!                                            tauyy = - mu_s*Syy      
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
  IF ( sgs_devsmag .EQ. 1 ) THEN
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                (r23*UYY(ij,1,k)-&
                r13*(UZZ(ij,1,k)+UXX(ij,1,k)))
        ENDDO
     ENDDO
  ELSE
     DO k=1,kmax
        DO ij=1,imax*jmax
           TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                UYY(ij,1,k)
        ENDDO
     ENDDO
  ENDIF
! Running statistics
  IF ( statflg .EQ. 1 ) THEN
     CALL LES_SAVE_AVGIJ(i6, dx, dy, dz, u, v, w,&
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d )
  ENDIF

! Dissipation, -tau_yy*du/dy
  IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, v_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
     DO k = 1,kmax
        DO ij = 1,imax*jmax
           h4(ij,1,k) =  h4(ij,1,k) - TAU_M(ij,1,k)*TMP_LOC(ij,1,k)
        ENDDO
     ENDDO
  ENDIF

! Calculate  - d/dy(tauyy)             (Y-momentum)
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, TAU_M(1,1,1), TMP_LOC(1,1,1), j1vsin, jmxvsin,&
       wrk1d, wrk2d, wrk3d)
  DO k=1,kmax
     DO ij=1,imax*jmax
        h2(ij,1,k) = h2(ij,1,k) - TMP_LOC(ij,1,k)
     ENDDO
  ENDDO
! Running statistics
  IF ( statflg .EQ. 1 ) THEN
     CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
          dy, TAU_M(1,1,1), TMP_LOC(1,1,1), j1vsin, jmxvsin,&
          wrk1d, wrk2d, wrk3d)
     CALL LES_SAVE_AVGIJ(i17, dx, dy, dz, u, v, w, &
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
          dx, TAU_M(1,1,1), TMP_LOC(1,1,1), i1vsin, imxvsin,&
          wrk1d, wrk2d, wrk3d) 
     CALL LES_SAVE_AVGIJ(i28, dx, dy, dz, u, v, w, &
          UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
  ENDIF

  IF ( kmax_total .GT. 1 ) THEN
! -----------------------------------------------------------------    
!                                            tauzz = - mu_s*Szz      
! !!! mu_s is in reality 2*mu_s !!!
! -----------------------------------------------------------------    
     IF ( sgs_devsmag .EQ. 1 ) THEN
        DO k=1,kmax
           DO ij=1,imax*jmax
              TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                   (r23*UZZ(ij,1,k)-&
                   r13*(UXX(ij,1,k)+UYY(ij,1,k)))
           ENDDO
        ENDDO
     ELSE
        DO k=1,kmax
           DO ij=1,imax*jmax
              TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                   UZZ(ij,1,k)
           ENDDO
        ENDDO
     ENDIF
! Running statistics
     IF ( statflg .EQ. 1 ) THEN
        IF ( sgs_devsmag .EQ. 1 ) THEN
           DO k=1,kmax
              DO ij=1,imax*jmax
                 TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                      (r23*UZZ(ij,1,k)-&
                      r13*(UXX(ij,1,k)+UYY(ij,1,k)))
              ENDDO
           ENDDO
        ELSE
           DO k=1,kmax
              DO ij=1,imax*jmax
                 TAU_M(ij,1,k) = -sgsvis(ij,1,k)*&
                      UZZ(ij,1,k)
              ENDDO
           ENDDO
        ENDIF
        CALL LES_SAVE_AVGIJ(i7, dx, dy, dz, u, v, w,&
             UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     ENDIF

! Dissipation, -tau_zz*du/dz
     IF ( iles_type_diss .EQ. LES_DISS_DISSSG ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, w_r, TMP_LOC(1,1,1), i0, i0, wrk1d, wrk2d, wrk3d)
        DO k = 1,kmax
           DO ij = 1,imax*jmax
              h4(ij,1,k) =  h4(ij,1,k) - TAU_M(ij,1,k)*TMP_LOC(ij,1,k)
           ENDDO
        ENDDO
     ENDIF

! Calculate  - d/dz(tauzz)             (Z-momentum)
     CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
          dz, TAU_M(1,1,1), TMP_LOC(1,1,1), k1vsin, kmxvsin,&
          wrk1d, wrk2d, wrk3d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           h3(ij,1,k) = h3(ij,1,k) - TMP_LOC(ij,1,k)
        ENDDO
     ENDDO
! Running statistics
     IF ( statflg .EQ. 1 ) THEN
        CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
             dz, TAU_M(1,1,1), TMP_LOC(1,1,1), k1vsin, kmxvsin,&
             wrk1d, wrk2d, wrk3d)
        CALL LES_SAVE_AVGIJ(i20, dx, dy, dz, u, v, w,&
             UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
        CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
             dx, TAU_M(1,1,1), UXX(1,1,1), i1vsin, imxvsin,&
             wrk1d, wrk2d, wrk3d)  
        CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
             dy, TAU_M(1,1,1), UYY(1,1,1), j1vsin, jmxvsin,&
             wrk1d, wrk2d, wrk3d) 
        CALL LES_SAVE_AVGIJ(i29, dx, dy, dz, u, v, w,&
             UXX(1,1,1), mean1d_les, wrk1d, wrk2d, wrk3d)
     ENDIF

  ENDIF

#undef SXY
#undef SXZ
#undef SYZ
#undef TAU_M
#undef TMP_LOC

#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'LEAVING SGS_SMG_U' )
#endif

  RETURN
END SUBROUTINE SGS_SMG_U
