!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2001/11/30 - J.P. Mellado
!#              Created
!#
!########################################################################
!# DESCRIPTION
!#
!# This subroutine computes the additive residual stress term of the right
!# hand side of the filtered compressible Navier-Stokes equations using a 
!# Smagorinsky model.
!#
!########################################################################
!# ARGUMENTS 
!#
!# txc    In   size 7
!#
!########################################################################
SUBROUTINE SGS_STA(x,dx,dy,dz, q,z1, h1,h2,h3,h4,zh1, mean1d_les, txc, chi, wrk1d,wrk2d)

#include "types.h"
#include "dns_const.h"
#include "les_avgij_map.h"

  USE DNS_GLOBAL
#ifdef PARALLEL
  USE DNS_MPI
#endif
   USE LES_GLOBAL

  IMPLICIT NONE

#include "integers.h"

  TREAL, DIMENSION(*)                :: x, dx, dy, dz
  TREAL, DIMENSION(imax,jmax,kmax)   :: h1, h2, h3, h4, chi
  TREAL, DIMENSION(imax,jmax,kmax,*) :: q, z1, zh1, txc
  TREAL, DIMENSION(nstatavg,jmax,*)  :: mean1d_les
  TREAL wrk1d(isize_wrk1d,*), wrk2d(*)

  TARGET :: q

! -----------------------------------------------------------------------
  TREAL nu_smin, nu_smax
  TREAL r23, r13, cs2, csmg
  TINTEGER ij, i, j, k, idsp

! Pointers to existing allocated space
  TREAL, DIMENSION(:,:,:),   POINTER :: u, v, w, e, rho

! #######################################################################
#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'ENTERING SGS_STA' )
#endif

! Define pointers
  u   => q(:,:,:,1)
  v   => q(:,:,:,2)
  w   => q(:,:,:,3)
  e   => q(:,:,:,4)
  rho => q(:,:,:,5)

  r23 = C_2_R/C_3_R
  r13 = C_1_R/C_3_R

! **********************
! *** txc(5) is S*S ***
! **********************

! ==============================
! Calculate (Sxx)**2 
! Sxx = 2/3 (du/dx) 
! ==============================
! ==============================
! Calculate (Syy)**2 
! Syy = 2/3 (dv/dy)
! ==============================
! ==============================
! Calculate (Szz)**2 
! Szz = 2/3 (dw/dz)
! ============================== 


  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, u, txc(1,1,1,1), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, v, txc(1,1,1,2), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))
  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, w, txc(1,1,1,3), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))

  IF ( sgs_devsmag .EQ. 1 ) THEN
     DO k=1,kmax
        DO ij=1,imax*jmax
           txc(ij,1,k,5) = (r23*txc(ij,1,k,1)-&
                r13*(txc(ij,1,k,2)+txc(ij,1,k,3)))**2 +&
                (r23*txc(ij,1,k,2)-&
                r13*(txc(ij,1,k,1)+txc(ij,1,k,3)))**2 +&
                (r23*txc(ij,1,k,3)-&
                r13*(txc(ij,1,k,2)+txc(ij,1,k,1)))**2
        ENDDO
     ENDDO
  ELSE
     DO k=1,kmax
        DO ij=1,imax*jmax
           txc(ij,1,k,5) = txc(ij,1,k,1)**2 +&
                txc(ij,1,k,2)**2 + txc(ij,1,k,3)**2
        ENDDO
     ENDDO
  ENDIF

! ============================================
! Calculate (Sxy)**2
! Sxy = 1/2 [ (du/dy) + (dv/dx) ]
! ============================================

  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, u, txc(1,1,1,2), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, v, txc(1,1,1,3), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))

  DO k=1,kmax
     DO ij=1,imax*jmax
        txc(ij,1,k,2) = C_05_R*(txc(ij,1,k,2)+txc(ij,1,k,3))
        txc(ij,1,k,5) = txc(ij,1,k,5) + C_2_R*txc(ij,1,k,2)**2
     ENDDO
  ENDDO

! ============================================
! Calculate (Sxz)**2
! Sxz = 1/2 [ (du/dz) + (dw/dx) ]
! ============================================

  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, u, txc(1,1,1,3), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))
  CALL PARTIAL_X(imode_fdm, imax, jmax, kmax, i1bc,&
       dx, w, txc(1,1,1,4), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))

  DO k=1,kmax
     DO ij=1,imax*jmax
        txc(ij,1,k,3) = C_05_R*(txc(ij,1,k,3)+txc(ij,1,k,4))
        txc(ij,1,k,5) = txc(ij,1,k,5) + C_2_R*txc(ij,1,k,3)**2
     ENDDO
  ENDDO

! ============================================
! Calculate (Syz)**2 
! Syz = 1/2 [ (dv/dz) + (dw/dy) ]
! ============================================      

  CALL PARTIAL_Z(imode_fdm, imax, jmax, kmax, k1bc,&
       dz, v, txc(1,1,1,4), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))
  CALL PARTIAL_Y(imode_fdm, imax, jmax, kmax, j1bc,&
       dy, w, txc(1,1,1,6), i0, i0,&
       wrk1d, wrk2d, txc(1,1,1,7))

  DO k=1,kmax
     DO ij=1,imax*jmax
        txc(ij,1,k,4) = C_05_R*(txc(ij,1,k,4)+txc(ij,1,k,6))
        txc(ij,1,k,5) = txc(ij,1,k,5) + C_2_R*txc(ij,1,k,4)**2
     ENDDO
  ENDDO

! ################################################################
! # txc(1) is rho*nu_s = rho*2*(Csm*sgs_delta)^2*SQRT(2*Sij*Sij) #
! ################################################################

#ifdef PARALLEL
  idsp = ims_offset_k
#else
  idsp = 0
#endif

  csmg = C_2_R*((sgs_csm*isgs_f0size)**2)*SQRT(C_2_R)

  IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
! Deactivation in the inflow zone for spatial mode
     IF ( sgs_smagdelta .EQ. C_0_R ) THEN
        DO i = 1,imax
           IF ( x(i) .LT. sgs_smagtrans ) THEN
              wrk1d(i,1) = C_0_R
           ELSE
              wrk1d(i,1) = C_1_R
           ENDIF
        ENDDO
     ELSE
        DO i = 1,imax
           wrk1d(i,1) = C_05_R*(C_1_R + &
                TANH(C_05_R*(x(i)-sgs_smagtrans)/sgs_smagdelta) )
        ENDDO
     ENDIF
#ifdef _DEBUG
     CALL IO_WRITE_ASCII_FIELD('smag-c0-x.dat', imax, i1, i1, wrk1d)
#endif
     DO k=1,kmax
        DO j=1, jmax
           DO i=1,imax
              cs2 = csmg*((dx(i)*dy(j)*dz(k+idsp))**r23)*wrk1d(i,1)
              txc(i,j,k,1) =  cs2 * SQRT(txc(i,j,k,5))
           ENDDO
        ENDDO
     ENDDO
  ELSE 
     DO k=1,kmax
        DO j=1, jmax
           DO i=1,imax
              cs2 = csmg*((dx(i)*dy(j)*dz(k+idsp))**r23)
              txc(i,j,k,1) =  cs2 * SQRT(txc(i,j,k,5))
           ENDDO
        ENDDO
     ENDDO
  ENDIF

  CALL MINMAX(imax,jmax,kmax, txc(1,1,1,1), nu_smin, nu_smax)
  smg_udiff = C_05_R*nu_smax

! Dynamic value
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        txc(ij,1,k,1) = rho(ij,1,k)*txc(ij,1,k,1)
     ENDDO
  ENDDO

! #####################################
! #   Momentum and energy equations   #
! #####################################
  CALL SGS_SMG_U(i0, i1, dx, dy, dz, rho, u, v, w, u, v, w,&
       txc(1,1,1,1), h1, h2, h3, h4, &
       txc(1,1,1,2), mean1d_les, wrk1d, wrk2d, txc(1,1,1,7))

! Calculate the dynamic diffusivity from SGS Schmidt number
  DO k=1,kmax
     DO ij=1,imax*jmax
        txc(ij,1,k,1) = C_05_R*txc(ij,1,k,1)/smg_prandtl
     ENDDO
  ENDDO
!  CALL SGS_SMG_P(i0, dx, dy, dz, rho, u, v, w, &
!       p, txc(1,1,1,1), h4, mean1d_les, txc(1,1,1,2), &
!       wrk1d, wrk2d, txc(1,1,1,7))
     CALL SGS_SMG_Z(i0, dx, dy, dz, rho, e, e, &
          txc(1,1,1,1), h4, mean1d_les, txc(1,1,1,2), chi, wrk1d, wrk2d, &
          txc(1,1,1,7))

! #####################################
! #         Scalar equation           #
! #####################################
  IF ( icalc_scal .EQ. 1 ) THEN
! Calculate the dynamic diffusivity from SGS Schmidt number
     DO k=1,kmax
        DO ij=1,imax*jmax
           txc(ij,1,k,1) = txc(ij,1,k,1)*smg_prandtl/smg_schmidt
        ENDDO
     ENDDO
     CALL SGS_SMG_Z(i0, dx, dy, dz, rho, z1, z1, &
          txc(1,1,1,1), zh1, mean1d_les(1+LA_MOMENTUM_SIZE*nstatavg*jmax,1,1), &
          txc(1,1,1,2), chi, wrk1d, wrk2d, &
          txc(1,1,1,7))
  ENDIF

#ifdef TRACE_ON
  CALL IO_WRITE_ASCII(tfile, 'LEAVING SGS_STA' )
#endif

  RETURN
END SUBROUTINE SGS_STA
