module mod_Traj
  use mod_data, only: type_numbering

  implicit none
  save

  double precision :: divergMax

  double precision :: jumpeps
  double precision :: pace
  double precision :: traj_err

  parameter(jumpeps=2.d-5)
  parameter(pace=0.02)
  parameter(traj_err=0.2d0)

  integer, parameter :: numendsmax=500000

  logical,                 dimension(:),      allocatable :: signs     
  double precision,        dimension(:,:),    allocatable :: points    
  integer(type_numbering), dimension(:,:,:,:),allocatable :: numbering 

contains

  !########################################################################
    subroutine initGlobalTraj()
      use mod_Data

      integer :: alloc_err=0

      allocate(signs(numendsmax),STAT=alloc_err)
      if( (alloc_err.NE.0) ) write(*,*) 'Not enough memory for *signs*. ERROR', alloc_err
      signs(:)=.true.
      
      allocate(points(3,numendsmax),STAT=alloc_err)
      if( (alloc_err.NE.0) ) write(*,*) 'Not enough memory for *points*. ERROR', alloc_err
      points(:,:)=0.0
      
      allocate(numbering(ndimx,ndimy,ndimz,2),STAT=alloc_err)
      if( (alloc_err.NE.0) ) write(*,*) 'Not enough memory for *numbering*. ERROR', alloc_err
      numbering(:,:,:,:)=0

      return
    end subroutine initGlobalTraj

  !########################################################################
    subroutine endGlobalTraj()
      use mod_Data

      deallocate( signs,points, numbering )
      
      return
    end subroutine endGlobalTraj

  !########################################################################
    subroutine loadConnectivity()
      use mod_Data
      
      open(21,file=filename,status='old',form='unformatted',access='stream')
      read(21) numends
      deallocate(points,signs)
      allocate(points(3,numends),signs(numends))

      read(21) numbering
      read(21) points
      read(21) signs
      close(21)

      return
    end subroutine loadConnectivity

  !########################################################################
    subroutine saveConnectivity()
      use mod_Data
      
      open(21,file=filename,form='unformatted',access='stream')
      write(21) numends
      write(21) numbering
      write(21) points(1:numends*3,1)
      write(21) signs(1:numends)
      close(21)

      return
    end subroutine saveConnectivity

  !########################################################################
  !# HISTORY
  !#
  !# 2007/11/20 - L. Wang
  !#              Created
  !# 2007/11/20 - J.H. Goebbert
  !#              F90 and OpenMP parallelized
  !# 2007/11/30 - J.P. Mellado
  !#              Adapted for free turbulent flows
  !#              Checking compact schemes
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !# Walks along the trajectory of every grid point in positive/negative direction
  !# until the minimum and maximum (or boundary) is found.
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine findTraj
    use omp_lib
    use mod_data
    use mod_DEle

    ! OpenMP private variables
    double precision :: switch
    integer          :: i,j,k, ni,nj,nk, dni,dnj,dnk, i0,j0,k0
    integer          :: force0, endfind

    double precision :: subdvx(-1:0,-1:1,-1:1)=0
    double precision :: subdvy(-1:1,-1:0,-1:1)=0
    double precision :: subdvz(-1:1,-1:1,-1:0)=0

    double precision :: endx,endy,endz
    double precision :: posx,posy,posz

    double precision :: traject(-10*ndimx:10*ndimx,11)
    integer          :: trajPtId, trajDir, trajStep
    integer          :: trajMaxMinIds(2)

    integer          :: jumps

    !########################################################################
    numends   = 0
    jumps     = 0
    ! set special values for 2D/3D. Here only 3D
    divergMax = -3.5

    write(*,*) 'Entering parallel section...'
    !$omp parallel                                                           &
    !$omp default( shared )                                                  &
    !$omp firstprivate( switch )                                             &
    !$omp firstprivate( i, j, k, ni, nj, nk, i0, j0, k0 )                    &
    !$omp firstprivate( force0, endfind )                                    &
    !$omp firstprivate( subdvx, subdvy, subdvz )                             &
    !$omp firstprivate( endx, endy, endz)                                    &
    !$omp firstprivate( posx, posy, posz )                                   &
    !$omp firstprivate( traject, trajPtId, trajDir, trajMaxMinIds, trajStep )&
    !$omp reduction(+:jumps)

    !$omp do schedule (dynamic,1)
    do i=subdomain(1,1),subdomain(1,2)
       write(*,*) 'i=',i
       do k=subdomain(3,1),subdomain(3,2)
          do j=subdomain(2,1),subdomain(2,2)
             
             ! calculation of fixed number of trajectories
             if ( itraj.and.itraj_stop ) goto 35

             !------------------------------------------------------------------------
             trajMaxMinIds(1) = 0        ! found points in positive direction
             trajMaxMinIds(2) = 0        ! found points in negative direction

             ! save values at grid-point as traject(0,:)
             endx = 0.d0
             endy = 0.d0
             endz = 0.d0
             if ( itraj ) call calcTrajPtValues(traject,0, endx,endy,endz, i,j,k)

             ! Trajectory - Calculation in positive and negative direction
             do trajDir=1,2

                if(trajDir.eq.1) then    ! positive direction
                   switch  =1.d0
                   trajStep=1
                else                     ! negative direction
                   switch  =-1.d0
                   trajStep=-1
                endif

                force0=0
                endfind=0

                posx=0
                posy=0
                posz=0

                ni=i
                nj=j
                nk=k

                trajPtId=0

                ! walk along the trajectory until a max/min is found
                do while(endfind.eq.0)

                   ! have we reached the boundary ?? exit loop if so
                   if(     (ni.eq.1).or.(ni.eq.ndimx) &
                       .or.(nk.eq.1).or.(nk.eq.ndimz) ) exit

                   ! Shear layer cropping
!                   if( (nj.lt.jmin_loc(ni,nk)).or.(nj.gt.jmax_loc(ni,nk)) ) then
!                      numbering(i,j,k,trajDir) = -1
!                      exit
!                   endif
                   if( nj.lt.jmin_loc(ni,nk) ) then ! lower boundary
                      numbering(i,j,k,trajDir) = -2
                      exit
                   endif
                   if( nj.gt.jmax_loc(ni,nk) ) then
                      numbering(i,j,k,trajDir) = -1 ! upper boundary
                      exit
                   endif
                  

                   ! interpolation of derivative
                   if ( ider_type .eq. 0 ) then ! second order
                      do i0=-1,0
                         do j0=-1,1
                            do k0=-1,1
                               subdvx(i0,j0,k0)= ( psi(ni+i0+1,nj+j0,nk+k0) &
                                    -psi(ni+i0,nj+j0,nk+k0) )/deltax
                            enddo
                         enddo
                      enddo
                      do i0=-1,1
                         do j0=-1,0
                            do k0=-1,1
                               subdvy(i0,j0,k0)= ( psi(ni+i0,nj+j0+1,nk+k0) &
                                    -psi(ni+i0,nj+j0,nk+k0) )/deltay
                            enddo
                         enddo
                      enddo
                      do i0=-1,1
                         do j0=-1,1
                            do k0=-1,0
                               subdvz(i0,j0,k0)= ( psi(ni+i0,nj+j0,nk+k0+1) &
                                    -psi(ni+i0,nj+j0,nk+k0) )/deltaz
                            enddo
                         enddo
                      enddo
                      
                   else ! compact
                      do i0=-1,0
                         do j0=-1,1
                            do k0=-1,1
                               subdvx(i0,j0,k0)= ( psi_x(ni+i0+1,nj+j0,nk+k0) &
                                    +psi_x(ni+i0,nj+j0,nk+k0) )/2.
                            enddo
                         enddo
                      enddo
                      do i0=-1,1
                         do j0=-1,0
                            do k0=-1,1
                               subdvy(i0,j0,k0)= ( psi_y(ni+i0,nj+j0+1,nk+k0) &
                                    +psi_y(ni+i0,nj+j0,nk+k0) )/2.
                            enddo
                         enddo
                      enddo
                      do i0=-1,1
                         do j0=-1,1
                            do k0=-1,0
                               subdvz(i0,j0,k0)= ( psi_z(ni+i0,nj+j0,nk+k0+1) &
                                    +psi_z(ni+i0,nj+j0,nk+k0) )/2.
                            enddo
                         enddo
                      enddo

                   endif

                   ! what is the path of trajectory through the grid-cell
                   call endpoint(posx, posy, posz, endx, endy, endz, switch, &
                        subdvx, subdvy, subdvz, endfind, jumps)

                   ! see thesis - stop infinite loop
                   force0=force0+1
                   if(force0.gt.500) then
                      endfind=1
                      exit
                   endif

                   ! last step to endpoint ?
                   if(endfind.eq.1) then
                      !$omp critical
                      ! Endpunktes in numbering,points
                      call numberfind(trajDir, i,j,k, numends, signs, points, numbering, &
                           ni+endx,nj+endy,nk+endz)
                      !$omp end critical
                   endif

                   ! calculate values on current trajectory-point
                   trajPtId = trajPtId+trajStep
                   if (itraj) call calcTrajPtValues(traject,trajPtId,endx,endy,endz,ni,nj,nk)
                   
                   ! check if trajectory left current staggered grid cell
                   ! call tellnew(ni,nj,nk,endx,endy,endz,posx,posy,posz)
                   dni=int(endx/0.5)
                   dnj=int(endy/0.5)
                   dnk=int(endz/0.5)
                   ni=ni+dni
                   nj=nj+dnj
                   nk=nk+dnk
                   posx=endx-dble(dni)
                   posy=endy-dble(dnj)
                   posz=endz-dble(dnk)

                enddo    ! endfind.eq.0

                ! save number of trajectory-points in negative and positive directions
                trajMaxMinIds(trajDir) = trajPtId

             enddo        ! trajDir

             ! analyse found trajectory: Jens' call to analyTrajStrain2PDF
             !$omp critical
             if (isavetraj) call storeDEleTraj(i,j,k,numbering, traject,trajMaxMinIds,traj_err)
             !$omp end critical

35        enddo            ! ndimz
       enddo               ! ndimy
    enddo                  ! ndimx

    !$omp end do
    write(*,*) 'thread: jumps =', jumps

    !$omp end parallel

    nj=0
    do i0 = 1,ndimx
       do k0 = 1,ndimz
          nj=nj+jmax_loc(i0,k0)-jmin_loc(i0,k0)+1
       enddo
    enddo
    write(*,*) 'jumps=',jumps
    write(*,*) 'total number points=',nj
    write(*,*) 'jumpratio=',jumps/dble(nj)
    write(*,*) 'numends=',numends

    return
  end subroutine findTraj

  !########################################################################
  !# HISTORY
  !#
  !# 2007/11/20 - J. H. Goebbert
  !#              Created
  !# 2007/11/30 - J.P. Mellado
  !#              Adding mapped fields
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !# Interpolates values on the trajectory at location nijk+endxyz and stores them
  !# in the the array 'traject(direction,stepNo,value)'
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine calcTrajPtValues(traject, trajPtId, endx,endy,endz, ni,nj,nk)
    use mod_Data

    double precision, intent(inout) :: traject(-10*ndimx:10*ndimx,*)
    integer,          intent(in)    :: trajPtId
    double precision, intent(in)    :: endx,endy,endz
    integer,          intent(in)    :: ni,nj,nk

    !------------------------------------------------------------------------
    integer          :: px,py,pz, fieldId
    double precision :: dx,dy,dz

    !########################################################################
    ! example 3.251431=>3    (-0.5<=endx=>0.5)
    px=ni+endx
    py=nj+endy
    pz=nk+endz

    dx=ni+endx-px
    dy=nj+endy-py
    dz=nk+endz-pz

    traject(trajPtId,1)=ni+endx
    traject(trajPtId,2)=nj+endy
    traject(trajPtId,3)=nk+endz

    ! conditioning field
    traject(trajPtId,4)=                                          &
          (1.d0-dx) *(1.d0-dy) *(1.d0-dz) *psi(px,  py,  pz)      &
         +dx        *(1.d0-dy) *(1.d0-dz) *psi(px+1,py,  pz)      &
         +(1.d0-dx) *dy        *(1.d0-dz) *psi(px,  py+1,pz)      &
         +(1.d0-dx) *(1.d0-dy) *dz        *psi(px,  py,  pz+1)    &
         +(1.d0-dx) *dy        *dz        *psi(px,  py+1,pz+1)    &
         +dx        *(1.d0-dy) *dz        *psi(px+1,py,  pz+1)    &
         +dx        *dy        *(1.d0-dz) *psi(px+1,py+1,pz)      &
         +dx        *dy        *dz        *psi(px+1,py+1,pz+1)

    ! additional mapped fields
    do fieldId = 1,dEleSets_noMapField
       traject(trajPtId,4+fieldId)=                                               &
             (1.d0-dx) *(1.d0-dy) *(1.d0-dz) *mappedfield(px,  py,  pz,  fieldId) &
            +dx        *(1.d0-dy) *(1.d0-dz) *mappedfield(px+1,py,  pz,  fieldId) &
            +(1.d0-dx) *dy        *(1.d0-dz) *mappedfield(px,  py+1,pz,  fieldId) &
            +(1.d0-dx) *(1.d0-dy) *dz        *mappedfield(px,  py,  pz+1,fieldId) &
            +(1.d0-dx) *dy        *dz        *mappedfield(px,  py+1,pz+1,fieldId) &
            +dx        *(1.d0-dy) *dz        *mappedfield(px+1,py,  pz+1,fieldId) &
            +dx        *dy        *(1.d0-dz) *mappedfield(px+1,py+1,pz,  fieldId) &
            +dx        *dy        *dz        *mappedfield(px+1,py+1,pz+1,fieldId)
    enddo

    return
  end subroutine calcTrajPtValues

  !########################################################################
  !# HISTORY
  !#
  !# 2007/11/20 - L. Wang
  !#              Created
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !# Stores information of minimum/maximum after trajectory-endpoint was found
  !# in 'numends', 'signs', 'points'
  !#
  !# Meaning of numbering(,,,2) is: 1.id of max point 2.id of min point
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine numberfind(idmm, i,j,k, numends,signs,points,numbering, tx,ty,tz)

    integer,                                                 intent(in)    :: idmm,i,j,k
    integer,                                                 intent(inout) :: numends
    logical,                 dimension(:),      allocatable, intent(inout) :: signs
    double precision,        dimension(:,:),    allocatable, intent(inout) :: points
    integer(type_numbering), dimension(:,:,:,:),allocatable, intent(inout) :: numbering
    double precision,                                        intent(in)    :: tx,ty,tz

    !------------------------------------------------------------------------
    double precision :: distance
    integer          :: endId

    !########################################################################
    do endId=1,numends
       distance = dsqrt((tx-points(1,endId))**2+(ty-points(2,endId))**2+(tz-points(3,endId))**2)
       if(distance.lt.traj_err) then
          numbering(i,j,k,idmm)=endId
          return
       endif
    enddo
    
    numends=numends+1
    
    numbering(i,j,k,idmm)=numends
    points(1,numends)=tx
    points(2,numends)=ty
    points(3,numends)=tz
    if(idmm.eq.1) then; signs(numends)=.true.
    else;               signs(numends)=.false.; endif
       
    return
  end subroutine numberfind

  !########################################################################
  !# HISTORY
  !#
  !# 2007/11/20 - L. Wang
  !#              Created
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !# Calculate end for a given pos. This routine works in relative local
  !# coordinates.
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine endpoint(posx,posy,posz, endx,endy,endz, switch, &
       subdvx,subdvy,subdvz, endfind, jumps)

    double precision,    intent(in)   :: posx,posy,posz
    double precision,    intent(out)  :: endx,endy,endz
    double precision,    intent(in)   :: switch
    double precision,    intent(in)   :: subdvx(-1:0,-1:1,-1:1)
    double precision,    intent(in)   :: subdvy(-1:1,-1:0,-1:1)
    double precision,    intent(in)   :: subdvz(-1:1,-1:1,-1:0)
    integer,             intent(inout):: endfind, jumps

    !------------------------------------------------------------------------
    double precision :: angle(3),temp,temp1,temp2,temp3
    double precision :: nposx,nposy,nposz
    double precision :: normx,normy,normz,normt,jump
    integer          :: ix,iy,iz,inner,force
    double precision :: xp,xm,yp,ym,zp,zm
    double precision :: vxp(3),vxm(3),vyp(3),vym(3),vzp(3),vzm(3)
    double precision :: diverg,totv,vsum(3)

    !########################################################################
    inner=1
    force=0

    nposx=posx
    nposy=posy
    nposz=posz

    normx=0.d0
    normy=0.d0
    normz=0.d0

    do ix=-1,0
       do iy=-1,1
          do iz=-1,1
             normx=max(dabs(subdvx(ix,iy,iz)),normx)
          enddo
       enddo
    enddo

    do ix=-1,1
       do iy=-1,0
          do iz=-1,1
             normy=max(dabs(subdvy(ix,iy,iz)),normy)
          enddo
       enddo
    enddo

    do ix=-1,1
       do iy=-1,1
          do iz=-1,0
             normz=max(dabs(subdvz(ix,iy,iz)),normz)
          enddo
       enddo
    enddo

    normt=dsqrt(normx*normx +normy*normy +normz*normz)+1.d-20

    do while(inner.eq.1)

       if((abs(nposx).gt.0.5d0).or.(abs(nposy).gt.0.5d0).or. (abs(nposz).gt.0.5d0)) then
          inner=0
          endx=nposx
          endy=nposy
          endz=nposz
          goto 54
       endif

       call findangle(nposx,nposy,nposz,switch,subdvx,subdvy,subdvz,angle,totv)

       jump=min(pace,totv/normt/4.d0)

       if(jump.lt.jumpeps) then

          xp=nposx+pace
          xm=nposx-pace
          yp=nposy+pace
          ym=nposy-pace
          zp=nposz+pace
          zm=nposz-pace
          call findangle(xp,nposy,nposz, switch, subdvx,subdvy,subdvz, vxp,temp)
          call findangle(xm,nposy,nposz, switch, subdvx,subdvy,subdvz, vxm,temp)
          call findangle(nposx,yp,nposz, switch, subdvx,subdvy,subdvz, vyp,temp)
          call findangle(nposx,ym,nposz, switch, subdvx,subdvy,subdvz, vym,temp)
          call findangle(nposx,nposy,zp, switch, subdvx,subdvy,subdvz, vzp,temp)
          call findangle(nposx,nposy,zm, switch, subdvx,subdvy,subdvz, vzm,temp)
          diverg=vxp(1)-vxm(1)+vyp(2)-vym(2)+vzp(3)-vzm(3)
          if(diverg.gt.divergMax) then
             call random_number(vsum(1))
             call random_number(vsum(2))
             call random_number(vsum(3))

             if(abs(angle(1)).ge.abs(angle(2)).and.abs(angle(1)).ge.abs(angle(3))) then
                temp2=vsum(2)**2*angle(2)
                temp3=vsum(3)**2*angle(3)
                temp1=10*(temp2*angle(2)+temp3*angle(3))/(angle(1)+1.d-20)
                temp=1.d-20+dsqrt(temp1**2+temp2**2+temp3**2)
                nposx=nposx-temp1/temp*pace
                nposy=nposy+temp2/temp*pace
                nposz=nposz+temp3/temp*pace
                goto 55
             else if(abs(angle(2)).ge.abs(angle(1)).and.abs(angle(2)).ge.abs(angle(3))) then
                temp1=vsum(1)**2*angle(1)
                temp3=vsum(3)**2*angle(3)
                temp2=10*(temp1*angle(1)+temp3*angle(3))/(angle(2)+1.d-20)
                temp=1.d-20+dsqrt(temp1**2+temp2**2+temp3**2)
                nposx=nposx+temp1/temp*pace
                nposy=nposy-temp2/temp*pace
                nposz=nposz+temp3/temp*pace
                goto 55
             else if(abs(angle(3)).ge.abs(angle(1)).and.abs(angle(3)).ge.abs(angle(2))) then
                temp1=vsum(1)**2*angle(1)
                temp2=vsum(2)**2*angle(2)
                temp3=10*(temp1*angle(1)+temp2*angle(2))/(angle(3)+1.d-20)
                temp=1.d-20+dsqrt(temp1**2+temp2**2+temp3**2)
                nposx=nposx+temp1/temp*pace
                nposy=nposy+temp2/temp*pace
                nposz=nposz-temp3/temp*pace
                goto 55
             endif
          endif

          endfind=1
          inner=0
          endx=nposx
          endy=nposy
          endz=nposz
          goto 54

       else
          nposx=nposx+angle(1)*jump
          nposy=nposy+angle(2)*jump
          nposz=nposz+angle(3)*jump
       endif

55     force=force+1
       if(force.gt.10000) then
          !write(*,*) 'this is a forced ending'
          jumps=jumps+1
          endfind=1
          inner=0
          endx=nposx
          endy=nposy
          endz=nposz
          goto 54
       endif
    enddo

54  return

  end subroutine endpoint

  !########################################################################
  !# HISTORY
  !#
  !# 2007/11/20 - L. Wang
  !#              Created
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine findangle(nposx, nposy, nposz, switch, subdvx, subdvy, subdvz, angl, totv)

    double precision,    intent(in)   :: nposx,nposy,nposz,switch
    double precision,    intent(in)   :: subdvx(-1:0,-1:1,-1:1)
    double precision,    intent(in)   :: subdvy(-1:1,-1:0,-1:1)
    double precision,    intent(in)   :: subdvz(-1:1,-1:1,-1:0)
    double precision,    intent(out)  :: angl(3)
    double precision,    intent(out)  :: totv

    !------------------------------------------------------------------------
    double precision :: sx,sy,sz
    integer          :: ix,iy,iz

    !########################################################################
    ix=-1
    iy=0
    iz=0
    sx=nposx+0.5d0
    sy=nposy
    sz=nposz

    if(nposy.lt.0.d0) then
       sy=1.d0+nposy
       iy=-1
    endif

    if(nposz.lt.0.d0) then
       sz=1.d0+nposz
       iz=-1
    endif

    angl(1)=    (1.d0-sx) *(1.d0-sy) *(1.d0-sz) *subdvx(ix,iy,iz)  &
         +sx        *(1.d0-sy) *(1.d0-sz) *subdvx(ix+1,iy,iz)      &
         +(1.d0-sx) *sy        *(1.d0-sz) *subdvx(ix,iy+1,iz)      &
         +(1.d0-sx) *(1.d0-sy) *sz        *subdvx(ix,iy,iz+1)      &
         +(1.d0-sx) *sy        *sz        *subdvx(ix,iy+1,iz+1)    &
         +sx        *(1.d0-sy) *sz        *subdvx(ix+1,iy,iz+1)    &
         +sx        *sy        *(1.d0-sz) *subdvx(ix+1,iy+1,iz)    &
         +sx        *sy        *sz        *subdvx(ix+1,iy+1,iz+1)
    angl(1)=angl(1)*switch

    ix=0
    iy=-1
    iz=0
    sx=nposx
    sy=nposy+0.5d0
    sz=nposz

    if(nposx.lt.0.d0) then
       sx=1.d0+nposx
       ix=-1
    endif

    if(nposz.lt.0.d0) then
       sz=1.d0+nposz
       iz=-1
    endif

    angl(2)=     (1.d0-sx) *(1.d0-sy) *(1.d0-sz) *subdvy(ix,iy,iz) &
         +sx        *(1.d0-sy) *(1.d0-sz) *subdvy(ix+1,iy,iz)      &
         +(1.d0-sx) *sy        *(1.d0-sz) *subdvy(ix,iy+1,iz)      &
         +(1.d0-sx) *(1.d0-sy) *sz        *subdvy(ix,iy,iz+1)      &
         +(1.d0-sx) *sy        *sz        *subdvy(ix,iy+1,iz+1)    &
         +sx        *(1.d0-sy) *sz        *subdvy(ix+1,iy,iz+1)    &
         +sx        *sy        *(1.d0-sz) *subdvy(ix+1,iy+1,iz)    &
         +sx        *sy        *sz        *subdvy(ix+1,iy+1,iz+1)
    angl(2)=angl(2)*switch

    ix=0
    iy=0
    iz=-1
    sx=nposx
    sy=nposy
    sz=nposz+0.5d0

    if(nposx.lt.0.d0) then
       sx=1.d0+nposx
       ix=-1
    endif

    if(nposy.lt.0.d0) then
       sy=1.d0+nposy
       iy=-1
    endif

    angl(3)=     (1.d0-sx) *(1.d0-sy) *(1.d0-sz) *subdvz(ix,iy,iz) &
         +sx        *(1.d0-sy) *(1.d0-sz) *subdvz(ix+1,iy,iz)      &
         +(1.d0-sx) *sy        *(1.d0-sz) *subdvz(ix,iy+1,iz)      &
         +(1.d0-sx) *(1.d0-sy) *sz        *subdvz(ix,iy,iz+1)      &
         +(1.d0-sx) *sy        *sz        *subdvz(ix,iy+1,iz+1)    &
         +sx        *(1.d0-sy) *sz        *subdvz(ix+1,iy,iz+1)    &
         +sx        *sy        *(1.d0-sz) *subdvz(ix+1,iy+1,iz)    &
         +sx        *sy        *sz        *subdvz(ix+1,iy+1,iz+1)
    angl(3)=angl(3)*switch

    totv=1.d-20+dsqrt(angl(1)**2+angl(2)**2+angl(3)**2)
    angl(1)=angl(1)/totv
    angl(2)=angl(2)/totv
    angl(3)=angl(3)/totv

    return
  end subroutine findangle

  !########################################################################
  !# HISTORY
  !#
  !# 2008/04/01 - J.P. Mellado
  !#              Routine findTraj adapted to use all of the data in the periodic cases
  !#
  !########################################################################
  !# DESCRIPTION
  !#
  !# Walks along the trajectory of every grid point in positive/negative direction
  !# until the minimum and maximum (or boundary) is found.
  !#
  !########################################################################
  !# ARGUMENTS 
  !#
  !########################################################################
  subroutine findTrajPeriodic
    use omp_lib
    use mod_data

    ! OpenMP private variables
    double precision :: switch
    integer          :: i,j,k, ni,nj,nk, dni,dnj,dnk, i0,j0,k0
    integer          :: force0, endfind
    integer          :: iper, iperp, iperm, kper, kperp, kperm

    double precision :: subdvx(-1:0,-1:1,-1:1)=0
    double precision :: subdvy(-1:1,-1:0,-1:1)=0
    double precision :: subdvz(-1:1,-1:1,-1:0)=0

    double precision :: endx,endy,endz, posx,posy,posz, tx,ty,tz

    integer           :: trajDir
    integer           :: jumps, forced_ends
    character(len=128):: str
    character(len=32) :: ofile='delements.log'

    !########################################################################
    numends     = 0
    jumps       = 0
    forced_ends = 0
    ! set special values for 2D/3D. Here only 3D
    divergMax   =-3.5

    !$omp parallel                                                           &
    !$omp default( shared )                                                  &
    !$omp firstprivate( switch )                                             &
    !$omp firstprivate( i, j, k, ni, nj, nk, i0, j0, k0 )                    &
    !$omp firstprivate( force0, endfind )                                    &
    !$omp firstprivate( subdvx, subdvy, subdvz )                             &
    !$omp firstprivate( endx, endy, endz)                                    &
    !$omp firstprivate( posx, posy, posz )                                   &
    !$omp firstprivate( str )                                                &
    !$omp reduction(+:jumps)                                                 &
    !$omp reduction(+:forced_ends)

    !$omp do schedule (dynamic,1)
    do i=subdomain(1,1),subdomain(1,2)
       WRITE(str,*) i; str='Iteration .....: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)

       do k=subdomain(3,1),subdomain(3,2)
          do j=subdomain(2,1),subdomain(2,2)
             
             endx = 0.d0; endy = 0.d0; endz = 0.d0

! Trajectory; calculation in positive and negative direction
             do trajDir=1,2

                if(trajDir.eq.1) then; switch= 1.d0             ! positive direction
                else;                  switch=-1.d0; endif      ! negative direction

                force0 =0
                endfind=0

                posx=0.d0; posy=0.d0; posz=0.d0
                ni=i;      nj=j;      nk=k

! Walk along the trajectory until a max/min is found
                do while(endfind.eq.0)
                   
                   ! Shear layer cropping
                   if( nj.lt.jmin_loc(ni,nk) ) then ! lower boundary
                      numbering(i,j,k,trajDir) = -2
                      exit
                   endif
                   if( nj.gt.jmax_loc(ni,nk) ) then ! upper boundary
                      numbering(i,j,k,trajDir) = -1 
                      exit
                   endif

                   ! Interpolation of derivative
                   do i0=-1,0
                      iperp=ni+i0+1; iperp=iperp+ndimx-1; iperp=mod(iperp,ndimx)+1
                      iperm=ni+i0;   iperm=iperm+ndimx-1; iperm=mod(iperm,ndimx)+1
                      do j0=-1,1; do k0=-1,1
                         kper=nk+k0; kper=kper+ndimz-1; kper=mod(kper,ndimz)+1
                         subdvx(i0,j0,k0)= ( psi(iperp,nj+j0,kper)-psi(iperm,nj+j0,kper) )/deltax
                      enddo; enddo
                   enddo
                   do i0=-1,1
                      iper=ni+i0; iper=iper+ndimx-1; iper=mod(iper,ndimx)+1
                      do j0=-1,0; do k0=-1,1
                         kper=nk+k0; kper=kper+ndimz-1; kper=mod(kper,ndimz)+1
                         subdvy(i0,j0,k0)= ( psi(iper,nj+j0+1,kper)-psi(iper,nj+j0,kper) )/deltay
                      enddo; enddo
                   enddo
                   do i0=-1,1
                      iper=ni+i0; iper=iper+ndimx-1; iper=mod(iper,ndimx)+1
                      do j0=-1,1;do k0=-1,0
                         kperp=nk+k0+1; kperp=kperp+ndimz-1; kperp=mod(kperp,ndimz)+1
                         kperm=nk+k0;   kperm=kperm+ndimz-1; kperm=mod(kperm,ndimz)+1
                         subdvz(i0,j0,k0)= ( psi(iper,nj+j0,kperp)-psi(iper,nj+j0,kperm) )/deltaz
                      enddo; enddo
                   enddo
                      
                   ! what is the path of trajectory through the grid-cell
                   call endpoint(posx,posy,posz, endx,endy,endz, switch, &
                        subdvx,subdvy,subdvz, endfind, jumps)
                   
                   ! Stop infinite loop. Maximum set to 2 x max. dimension
                   force0=force0+1
                   if ( force0 .gt. 2*max(ndimx,ndimy,ndimz) ) then
                      endfind=1
                      forced_ends = forced_ends + 1
                   endif

                   ! last step to endpoint ?
                   if(endfind.eq.1) then
                      !$omp critical
                      ! Endpunktes in numbering,points
                      tx=dble(ni)+endx; if (tx.lt.1.0) tx=tx+dble(ndimx)
                      ty=dble(nj)+endy
                      tz=dble(nk)+endz; if (tz.lt.1.0) tz=tz+dble(ndimz)
                      call numberfind(trajDir, i,j,k, numends, signs,points,numbering, tx,ty,tz)
                      !$omp end critical
                   endif

                   ! new reference node and updated value of pos
                   dni=int(endx/0.5d0); dnj=int(endy/0.5d0); dnk=int(endz/0.5d0)
                   ni=ni+dni; ni=ni+ndimx-1; ni=mod(ni,ndimx)+1
                   nj=nj+dnj
                   nk=nk+dnk; nk=nk+ndimz-1; nk=mod(nk,ndimz)+1
                   posx=endx-dble(dni); posy=endy-dble(dnj); posz=endz-dble(dnk)

                enddo      ! endfind.eq.0

             enddo         ! trajDir

          enddo            ! ndimz
       enddo               ! ndimy
    enddo                  ! ndimx

    !$omp end do

    WRITE(str,*) forced_ends; str = 'thread number of forced ends : '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)
    WRITE(str,*) jumps;       str = 'thread number of jumps ......: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)

    !$omp end parallel

    nj=0
    do i0=1,ndimx; do k0 =1,ndimz
       nj=nj+jmax_loc(i0,k0)-jmin_loc(i0,k0)+1
    enddo; enddo
    WRITE(str,*) nj;          str = 'number points .............: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)
    WRITE(str,*) numends;     str = 'number extrema ............: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)
    WRITE(str,*) forced_ends; str = 'number of forced ends .....: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)
    WRITE(str,*) jumps;       str = 'number of jumps ...........: '//TRIM(ADJUSTL(str)); CALL IO_WRITE_ASCII(ofile,str)

    return
  end subroutine findTrajPeriodic

end module mod_Traj
